@@ -1,22 +1,32 @@
 import copy
-import sys
 import itertools
-from typing import List, Optional
+import logging
+from typing import Optional
 
-from pyvex.expr import IRExpr
-from pyvex.stmt import IRStmt
-
-from . import VEXObject
 from . import expr, stmt
 from .const import get_type_size
-from .stmt import WrTmp, LoadG, LLSC, Dirty, CAS, get_enum_from_int, get_int_from_enum, Exit, IMark
-from .expr import RdTmp
+from .const_val import ConstVal
 from .data_ref import DataRef
+from .enums import VEXObject
 from .errors import SkipStatementsError
+from .expr import RdTmp
+from .native import pvc
+from .stmt import (
+    CAS,
+    LLSC,
+    Dirty,
+    Exit,
+    IMark,
+    IRExpr,
+    IRStmt,
+    LoadG,
+    WrTmp,
+    get_enum_from_int,
+    get_int_from_enum,
+)
+from .types import Arch
 
-
-import logging
-l = logging.getLogger("pyvex.block")
+log = logging.getLogger("pyvex.block")
 
 
 class IRSB(VEXObject):
@@ -27,8 +37,7 @@ class IRSB(VEXObject):
     IRSB stands for *Intermediate Representation Super-Block*. An IRSB in VEX is a single-entry, multiple-exit code
     block.
 
-    :ivar arch:             The architecture this block is lifted under
-    :vartype arch:          :class:`archinfo.Arch`
+    :ivar arch:             The architecture this block is lifted under. Must duck-type as :class:`archinfo.arch.Arch`
     :ivar statements:       The statements in this block
     :vartype statements:    list of :class:`IRStmt`
     :ivar next:             The expression for the default exit target of this block
@@ -41,27 +50,57 @@ class IRSB(VEXObject):
     :ivar int addr:         The address of this basic block, i.e. the address in the first IMark
     """
 
-    __slots__ = ('addr', 'arch', 'statements', 'next', '_tyenv', 'jumpkind', '_direct_next', '_size', '_instructions',
-                 '_exit_statements', 'default_exit_target', '_instruction_addresses', 'data_refs', )
+    __slots__ = (
+        "addr",
+        "arch",
+        "statements",
+        "next",
+        "_tyenv",
+        "jumpkind",
+        "_direct_next",
+        "_size",
+        "_instructions",
+        "_exit_statements",
+        "default_exit_target",
+        "_instruction_addresses",
+        "data_refs",
+        "const_vals",
+    )
 
     # The following constants shall match the defs in pyvex.h
     MAX_EXITS = 400
     MAX_DATA_REFS = 2000
+    MAX_CONST_VALS = 1000
 
-    def __init__(self, data, mem_addr, arch, max_inst=None, max_bytes=None,
-                 bytes_offset=0, traceflags=0, opt_level=1, num_inst=None, num_bytes=None, strict_block_end=False,
-                 skip_stmts=False, collect_data_refs=False, cross_insn_opt=True):
+    def __init__(
+        self,
+        data,
+        mem_addr,
+        arch: Arch,
+        max_inst=None,
+        max_bytes=None,
+        bytes_offset=0,
+        traceflags=0,
+        opt_level=1,
+        num_inst=None,
+        num_bytes=None,
+        strict_block_end=False,
+        skip_stmts=False,
+        collect_data_refs=False,
+        cross_insn_opt=True,
+    ):
         """
         :param data:                The bytes to lift. Can be either a string of bytes or a cffi buffer object.
                                     You may also pass None to initialize an empty IRSB.
         :type data:                 str or bytes or cffi.FFI.CData or None
         :param int mem_addr:        The address to lift the data at.
         :param arch:                The architecture to lift the data as.
-        :type arch:                 :class:`archinfo.Arch`
         :param max_inst:            The maximum number of instructions to lift. (See note below)
         :param max_bytes:           The maximum number of bytes to use.
-        :param num_inst:            Replaces max_inst if max_inst is None. If set to None as well, no instruction limit is used.
-        :param num_bytes:           Replaces max_bytes if max_bytes is None. If set to None as well, no  byte limit is used.
+        :param num_inst:            Replaces max_inst if max_inst is None. If set to None as well, no instruction limit
+                                    is used.
+        :param num_bytes:           Replaces max_bytes if max_bytes is None. If set to None as well, no  byte limit is
+                                    used.
         :param bytes_offset:        The offset into `data` to start lifting at. Note that for ARM THUMB mode, both
                                     `mem_addr` and `bytes_offset` must be odd (typically `bytes_offset` is set to 1).
         :param traceflags:          The libVEX traceflags, controlling VEX debug prints.
@@ -81,39 +120,47 @@ class IRSB(VEXObject):
         .. note:: If no instruction and byte limit is used, pyvex will continue lifting the block until the block
                   ends properly or until it runs out of data to lift.
         """
-        if max_inst is None: max_inst = num_inst
-        if max_bytes is None: max_bytes = num_bytes
+        if max_inst is None:
+            max_inst = num_inst
+        if max_bytes is None:
+            max_bytes = num_bytes
         VEXObject.__init__(self)
         self.addr = mem_addr
-        self.arch = arch
+        self.arch: Arch = arch
 
-        self.statements = [] # type: List[IRStmt]
-        self.next = None # type: Optional[IRExpr]
-        self._tyenv = None
-        self.jumpkind = None # type: Optional[str]
-        self._direct_next = None
-        self._size = None
-        self._instructions = None
-        self._exit_statements = None
+        self.statements: list[IRStmt] = []
+        self.next: IRExpr | None = None
+        self._tyenv: Optional["IRTypeEnv"] = None
+        self.jumpkind: str | None = None
+        self._direct_next: bool | None = None
+        self._size: int | None = None
+        self._instructions: int | None = None
+        self._exit_statements: tuple[tuple[int, int, IRStmt], ...] | None = None
         self.default_exit_target = None
         self.data_refs = ()
-        self._instruction_addresses = ()
+        self.const_vals = ()
+        self._instruction_addresses: tuple[int, ...] = ()
 
         if data is not None:
             # This is the slower path (because we need to call _from_py() to copy the content in the returned IRSB to
             # the current IRSB instance. You should always call `lift()` directly. This method is kept for compatibility
             # concerns.
-            irsb = lift(data, mem_addr, arch,
-                        max_bytes=max_bytes,
-                        max_inst=max_inst,
-                        bytes_offset=bytes_offset,
-                        opt_level=opt_level,
-                        traceflags=traceflags,
-                        strict_block_end=strict_block_end,
-                        skip_stmts=skip_stmts,
-                        collect_data_refs=collect_data_refs,
-                        cross_insn_opt=cross_insn_opt,
-                        )
+            from pyvex.lifting import lift
+
+            irsb = lift(
+                data,
+                mem_addr,
+                arch,
+                max_bytes=max_bytes,
+                max_inst=max_inst,
+                bytes_offset=bytes_offset,
+                opt_level=opt_level,
+                traceflags=traceflags,
+                strict_block_end=strict_block_end,
+                skip_stmts=skip_stmts,
+                collect_data_refs=collect_data_refs,
+                cross_insn_opt=cross_insn_opt,
+            )
             self._from_py(irsb)
 
     @staticmethod
@@ -123,7 +170,7 @@ class IRSB(VEXObject):
         return block
 
     @property
-    def tyenv(self):
+    def tyenv(self) -> "IRTypeEnv":
         if self._tyenv is None:
             self._tyenv = IRTypeEnv(self.arch)
         return self._tyenv
@@ -133,35 +180,35 @@ class IRSB(VEXObject):
         self._tyenv = v
 
     @property
-    def has_statements(self):
-        return self.statements is not None and self.statements
+    def has_statements(self) -> bool:
+        return self.statements is not None and bool(self.statements)
 
     @property
-    def exit_statements(self):
-
+    def exit_statements(self) -> tuple[tuple[int, int, IRStmt], ...]:
         if self._exit_statements is not None:
             return self._exit_statements
 
         # Delayed process
         if not self.has_statements:
-            return [ ]
+            return ()
 
-        self._exit_statements = [ ]
+        exit_statements = []
 
         ins_addr = None
         for idx, stmt_ in enumerate(self.statements):
             if type(stmt_) is IMark:
                 ins_addr = stmt_.addr + stmt_.delta
             elif type(stmt_) is Exit:
-                self._exit_statements.append((ins_addr, idx, stmt_))
+                assert ins_addr is not None
+                exit_statements.append((ins_addr, idx, stmt_))
 
-        self._exit_statements = tuple(self._exit_statements)
+        self._exit_statements = tuple(exit_statements)
         return self._exit_statements
 
-    def copy(self):
+    def copy(self) -> "IRSB":
         return copy.deepcopy(self)
 
-    def extend(self, extendwith):
+    def extend(self, extendwith) -> None:
         """
         Appends an irsb to the current irsb. The irsb that is appended is invalidated. The appended irsb's jumpkind and
         default exit are used.
@@ -172,8 +219,8 @@ class IRSB(VEXObject):
             self._from_py(extendwith)
             return
 
-        conversion_dict = { }
-        invalid_vals = (0xffffffff, -1)
+        conversion_dict = {}
+        invalid_vals = (0xFFFFFFFF, -1)
 
         new_size = self.size + extendwith.size
         new_instructions = self.instructions + extendwith.instructions
@@ -192,8 +239,8 @@ class IRSB(VEXObject):
 
         def convert_expr(expr_):
             """
-            Converts a VEX expression to use tmps in the appended-block instead of the appended-to-block. Used to prevent
-            collisions in tmp numbers between the two blocks.
+            Converts a VEX expression to use tmps in the appended-block instead of the appended-to-block. Used to
+            prevent collisions in tmp numbers between the two blocks.
             :param tmp:       The VEX expression to convert
             :vartype expr:    :class:`IRExpr`
             """
@@ -215,10 +262,12 @@ class IRSB(VEXObject):
                 for e in stmt_.args:
                     convert_expr(e)
             elif stmttype is CAS:
-                if stmt_.oldLo not in invalid_vals: stmt_.oldLo = convert_tmp(stmt_.oldLo)
-                if stmt_.oldHi not in invalid_vals: stmt_.oldHi = convert_tmp(stmt_.oldHi)
+                if stmt_.oldLo not in invalid_vals:
+                    stmt_.oldLo = convert_tmp(stmt_.oldLo)
+                if stmt_.oldHi not in invalid_vals:
+                    stmt_.oldHi = convert_tmp(stmt_.oldHi)
             # Convert all expressions
-            to_replace = { }
+            to_replace = {}
             for expr_ in stmt_.expressions:
                 replacement = convert_expr(expr_)
                 if replacement is not expr_:
@@ -235,43 +284,44 @@ class IRSB(VEXObject):
 
         # TODO: Change exit_statements, data_references, etc.
 
-    def invalidate_direct_next(self):
+    def invalidate_direct_next(self) -> None:
         self._direct_next = None
 
-    def pp(self):
+    def pp(self) -> None:
         """
         Pretty-print the IRSB to stdout.
         """
         print(self._pp_str())
 
     def __repr__(self):
-        return 'IRSB <0x%x bytes, %s ins., %s> at 0x%x' % (self.size, self.instructions, str(self.arch), self.addr)
+        return f"IRSB <0x{self.size:x} bytes, {self.instructions} ins., {str(self.arch)}> at 0x{self.addr:x}"
 
     def __str__(self):
         return self._pp_str()
 
     def __eq__(self, other):
-        return (isinstance(other, IRSB) and
-                self.addr == other.addr and
-                self.arch.name == other.arch.name and
-                self.statements == other.statements and
-                self.next == other.next and
-                self.jumpkind == other.jumpkind
-                )
+        return (
+            isinstance(other, IRSB)
+            and self.addr == other.addr
+            and self.arch.name == other.arch.name
+            and self.statements == other.statements
+            and self.next == other.next
+            and self.jumpkind == other.jumpkind
+        )
 
     def __hash__(self):
         return hash((IRSB, self.addr, self.arch.name, tuple(self.statements), self.next, self.jumpkind))
 
-    def typecheck(self):
+    def typecheck(self) -> bool:
         try:
             # existence assertions
             assert self.next is not None, "Missing next expression"
             assert self.jumpkind is not None, "Missing jumpkind"
 
-            # type assertions
+            # Type assertions
             assert isinstance(self.next, expr.IRExpr), "Next expression is not an expression"
             assert type(self.jumpkind is str), "Jumpkind is not a string"
-            assert self.jumpkind.startswith('Ijk_'), "Jumpkind is not a jumpkind enum"
+            assert self.jumpkind.startswith("Ijk_"), "Jumpkind is not a jumpkind enum"
             assert self.tyenv.typecheck(), "Type environment contains invalid types"
 
             # statement assertions
@@ -280,7 +330,7 @@ class IRSB(VEXObject):
                 assert isinstance(st, stmt.IRStmt), "Statement %d is not an IRStmt" % i
                 try:
                     assert st.typecheck(self.tyenv), "Statement %d failed to typecheck" % i
-                except: # pylint: disable=bare-except
+                except Exception:  # pylint: disable=bare-except
                     assert False, "Statement %d errored in typechecking" % i
 
                 if type(st) is stmt.NoOp:
@@ -295,7 +345,7 @@ class IRSB(VEXObject):
 
             assert last_imark is not None, "No IMarks present in block"
         except AssertionError as e:
-            l.debug(e.args[0])
+            log.debug(e.args[0])
             return False
         return True
 
@@ -304,13 +354,13 @@ class IRSB(VEXObject):
     #
 
     @staticmethod
-    def from_c(c_irsb, mem_addr, arch):
+    def from_c(c_irsb, mem_addr, arch) -> "IRSB":
         irsb = IRSB(None, mem_addr, arch)
         irsb._from_c(c_irsb)
         return irsb
 
     @staticmethod
-    def from_py(tyenv, stmts, next_expr, jumpkind, mem_addr, arch):
+    def from_py(tyenv, stmts, next_expr, jumpkind, mem_addr, arch) -> "IRSB":
         irsb = IRSB(None, mem_addr, arch)
 
         irsb.tyenv = tyenv
@@ -326,13 +376,13 @@ class IRSB(VEXObject):
     #
 
     @property
-    def stmts_used(self):
+    def stmts_used(self) -> int:
         if self.statements is None:
             return 0
         return len(self.statements)
 
     @property
-    def offsIP(self):
+    def offsIP(self) -> int:
         return self.arch.ip_offset
 
     @property
@@ -347,8 +397,7 @@ class IRSB(VEXObject):
         Return an iterator of all expressions contained in the IRSB.
         """
         for s in self.statements:
-            for expr_ in s.expressions:
-                yield expr_
+            yield from s.expressions
         yield self.next
 
     @property
@@ -364,15 +413,17 @@ class IRSB(VEXObject):
         return self._instructions
 
     @property
-    def instruction_addresses(self):
+    def instruction_addresses(self) -> tuple[int, ...]:
         """
         Addresses of instructions in this block.
         """
         if self._instruction_addresses is None:
             if self.statements is None:
-                self._instruction_addresses = [ ]
+                self._instruction_addresses = ()
             else:
-                self._instruction_addresses = [ (s.addr + s.delta) for s in self.statements if type(s) is stmt.IMark ]
+                self._instruction_addresses = tuple(
+                    (s.addr + s.delta) for s in self.statements if type(s) is stmt.IMark
+                )
         return self._instruction_addresses
 
     @property
@@ -391,14 +442,15 @@ class IRSB(VEXObject):
         """
         ops = []
         for e in self.expressions:
-            if hasattr(e, 'op'):
+            if hasattr(e, "op"):
                 ops.append(e.op)
         return ops
 
     @property
     def all_constants(self):
         """
-        Returns all constants in the block (including incrementing of the program counter) as :class:`pyvex.const.IRConst`.
+        Returns all constants in the block (including incrementing of the program counter) as
+        :class:`pyvex.const.IRConst`.
         """
         return sum((e.constants for e in self.expressions), [])
 
@@ -407,8 +459,7 @@ class IRSB(VEXObject):
         """
         The constants (excluding updates of the program counter) in the IRSB as :class:`pyvex.const.IRConst`.
         """
-        return sum(
-            (s.constants for s in self.statements if not (type(s) is stmt.Put and s.offset == self.offsIP)), [])
+        return sum((s.constants for s in self.statements if not (type(s) is stmt.Put and s.offset == self.offsIP)), [])
 
     @property
     def constant_jump_targets(self):
@@ -432,7 +483,7 @@ class IRSB(VEXObject):
         """
         A dict of the static jump targets of the basic block to their jumpkind.
         """
-        exits = dict()
+        exits = {}
 
         if self.exit_statements:
             for _, _, stmt_ in self.exit_statements:
@@ -448,11 +499,9 @@ class IRSB(VEXObject):
     # private methods
     #
 
-    def _pp_str(self):
+    def _pp_str(self) -> str:
         """
         Return the pretty-printed IRSB.
-
-        :rtype: str
         """
         sa = []
         sa.append("IRSB {")
@@ -462,26 +511,29 @@ class IRSB(VEXObject):
         if self.statements is not None:
             for i, s in enumerate(self.statements):
                 if isinstance(s, stmt.Put):
-                    stmt_str = s.__str__(reg_name=self.arch.translate_register_name(s.offset, s.data.result_size(self.tyenv) // 8))
+                    stmt_str = s.pp_str(
+                        reg_name=self.arch.translate_register_name(s.offset, s.data.result_size(self.tyenv) // 8)
+                    )
                 elif isinstance(s, stmt.WrTmp) and isinstance(s.data, expr.Get):
-                    stmt_str = s.__str__(reg_name=self.arch.translate_register_name(s.data.offset, s.data.result_size(self.tyenv) // 8))
+                    stmt_str = s.pp_str(
+                        reg_name=self.arch.translate_register_name(s.data.offset, s.data.result_size(self.tyenv) // 8)
+                    )
                 elif isinstance(s, stmt.Exit):
-                    stmt_str = s.__str__(reg_name=self.arch.translate_register_name(s.offsIP, self.arch.bits // 8))
+                    stmt_str = s.pp_str(reg_name=self.arch.translate_register_name(s.offsIP, self.arch.bits // 8))
                 else:
-                    stmt_str = s.__str__()
+                    stmt_str = s.pp_str()
                 sa.append("   %02d | %s" % (i, stmt_str))
         else:
             sa.append("   Statements are omitted.")
-        sa.append(
-            "   NEXT: PUT(%s) = %s; %s" % (self.arch.translate_register_name(self.offsIP), self.next, self.jumpkind))
+        sa.append(f"   NEXT: PUT({self.arch.translate_register_name(self.offsIP)}) = {self.next}; {self.jumpkind}")
         sa.append("}")
-        return '\n'.join(sa)
+        return "\n".join(sa)
 
     def _is_defaultexit_direct_jump(self):
         """
         Checks if the default of this IRSB a direct jump or not.
         """
-        if not (self.jumpkind == 'Ijk_InvalICache' or self.jumpkind == 'Ijk_Boring' or self.jumpkind == 'Ijk_Call'):
+        if not (self.jumpkind == "Ijk_InvalICache" or self.jumpkind == "Ijk_Boring" or self.jumpkind == "Ijk_Call"):
             return False
 
         target = self.default_exit_target
@@ -507,7 +559,7 @@ class IRSB(VEXObject):
         self._instruction_addresses = tuple(itertools.islice(lift_r.inst_addrs, lift_r.insts))
 
         # Conditional exits
-        self._exit_statements = [ ]
+        exit_statements = []
         if skip_stmts:
             if lift_r.exit_count > self.MAX_EXITS:
                 # There are more exits than the default size of the exits array. We will need all statements
@@ -515,9 +567,9 @@ class IRSB(VEXObject):
             for i in range(lift_r.exit_count):
                 ex = lift_r.exits[i]
                 exit_stmt = stmt.IRStmt._from_c(ex.stmt)
-                self._exit_statements.append((ex.ins_addr, ex.stmt_idx, exit_stmt))
+                exit_statements.append((ex.ins_addr, ex.stmt_idx, exit_stmt))
 
-            self._exit_statements = tuple(self._exit_statements)
+            self._exit_statements = tuple(exit_statements)
         else:
             self._exit_statements = None  # It will be generated when self.exit_statements is called
         # The default exit
@@ -530,11 +582,29 @@ class IRSB(VEXObject):
         self.data_refs = None
         if lift_r.data_ref_count > 0:
             if lift_r.data_ref_count > self.MAX_DATA_REFS:
-                raise SkipStatementsError("data_ref_count exceeded MAX_DATA_REFS (%d)" % self.MAX_DATA_REFS)
+                raise SkipStatementsError(f"data_ref_count exceeded MAX_DATA_REFS ({self.MAX_DATA_REFS})")
             self.data_refs = [DataRef.from_c(lift_r.data_refs[i]) for i in range(lift_r.data_ref_count)]
 
-    def _set_attributes(self, statements=None, nxt=None, tyenv=None, jumpkind=None, direct_next=None, size=None,
-                        instructions=None, instruction_addresses=None, exit_statements=None, default_exit_target=None):
+        # Const values
+        self.const_vals = None
+        if lift_r.const_val_count > 0:
+            if lift_r.const_val_count > self.MAX_CONST_VALS:
+                raise SkipStatementsError(f"const_val_count exceeded MAX_CONST_VALS ({self.MAX_CONST_VALS})")
+            self.const_vals = [ConstVal.from_c(lift_r.const_vals[i]) for i in range(lift_r.const_val_count)]
+
+    def _set_attributes(
+        self,
+        statements=None,
+        nxt=None,
+        tyenv=None,
+        jumpkind=None,
+        direct_next=None,
+        size=None,
+        instructions=None,
+        instruction_addresses=None,
+        exit_statements=None,
+        default_exit_target=None,
+    ):
         self.statements = statements if statements is not None else []
         self.next = nxt
         if tyenv is not None:
@@ -548,10 +618,19 @@ class IRSB(VEXObject):
         self.default_exit_target = default_exit_target
 
     def _from_py(self, irsb):
-        self._set_attributes(irsb.statements, irsb.next, irsb.tyenv, irsb.jumpkind, irsb.direct_next, irsb.size,
-                             instructions=irsb._instructions, instruction_addresses=irsb._instruction_addresses,
-                             exit_statements=irsb.exit_statements, default_exit_target=irsb.default_exit_target,
-                             )
+        self._set_attributes(
+            irsb.statements,
+            irsb.next,
+            irsb.tyenv,
+            irsb.jumpkind,
+            irsb.direct_next,
+            irsb.size,
+            instructions=irsb._instructions,
+            instruction_addresses=irsb._instruction_addresses,
+            exit_statements=irsb.exit_statements,
+            default_exit_target=irsb.default_exit_target,
+        )
+
 
 class IRTypeEnv(VEXObject):
     """
@@ -562,22 +641,22 @@ class IRTypeEnv(VEXObject):
     :vartype types:     list of str
     """
 
-    __slots__ = ['types', 'wordty']
+    __slots__ = ["types", "wordty"]
 
     def __init__(self, arch, types=None):
         VEXObject.__init__(self)
         self.types = [] if types is None else types
-        self.wordty = 'Ity_I%d' % arch.bits
+        self.wordty = "Ity_I%d" % arch.bits
 
     def __str__(self):
-        return ' '.join(("t%d:%s" % (i, t)) for i, t in enumerate(self.types))
+        return " ".join(("t%d:%s" % (i, t)) for i, t in enumerate(self.types))
 
     def lookup(self, tmp):
         """
         Return the type of temporary variable `tmp` as an enum string
         """
         if tmp < 0 or tmp > self.types_used:
-            l.debug("Invalid temporary number %d", tmp)
+            log.debug("Invalid temporary number %d", tmp)
             raise IndexError(tmp)
         return self.types[tmp]
 
@@ -613,6 +692,3 @@ class IRTypeEnv(VEXObject):
             except ValueError:
                 return False
         return True
-
-from . import pvc
-from .lifting import lift
